"""Defines the models for the different forms of Activities that can be scheduled."""
from .Action import Action
from .ActionType import ActionType
from .CallActivityAction import CallActivityAction
from .CallActivityWithRetryAction import CallActivityWithRetryAction
from .WaitForExternalEventAction import WaitForExternalEventAction
from .CallHttpAction import CallHttpAction
from .CreateTimerAction import CreateTimerAction

__all__ = [
    'Action',
    'ActionType',
    'CallActivityAction',
    'CallActivityWithRetryAction',
    'CallHttpAction',
    'WaitForExternalEventAction',
    'CreateTimerAction'
]



import json
import logging

logger = logging.getLogger('activity')


def main(value: str):
	logger.warn(f'ReturnListOfFloat is called with {value}')
	float_value = float(value)
	result = []
	result.append(float_value)
	result.append(float_value + 1.0)
	result.append(float_value + 2.0)
	return result



from typing import List
from ..models.actions.CreateTimerAction import CreateTimerAction
from ..models.history import HistoryEvent
from .task_utilities import find_task_timer_created, find_task_retry_timer_fired, set_processed
import datetime
from .timer_task import TimerTask


def create_timer_task(state: List[HistoryEvent],
                      fire_at: datetime) -> TimerTask:
    """Durable Timers are used in orchestrator function to implement delays.

    Parameters
    ----------
    state : List[HistoryEvent]
        The list of history events to search to determine the current state of the activity
    fire_at : datetime
        The time interval to fire the timer trigger

    Returns
    -------
    TimerTask
        A Durable Timer Task that schedules the timer to wake up the activity
    """
    new_action = CreateTimerAction(fire_at)

    timer_created = find_task_timer_created(state, fire_at)
    timer_fired = find_task_retry_timer_fired(state, timer_created)

    set_processed([timer_created, timer_fired])

    if timer_fired:
        return TimerTask(
            is_completed=True, action=new_action,
            timestamp=timer_fired.timestamp,
            id_=timer_fired.event_id)
    else:
        return TimerTask(
            is_completed=False, action=new_action,
            timestamp=None,
            id_=None)



"""Contains the definitions for the functions that enable scheduling of activities."""
from .call_activity import call_activity_task
from .call_activity_with_retry import call_activity_with_retry_task
from .task_all import task_all
from .task_any import task_any
from .task_utilities import should_suspend
from .wait_for_external_event import wait_for_external_event_task
from .continue_as_new import continue_as_new
from .new_uuid import new_uuid
from .call_http import call_http
from .create_timer import create_timer_task

__all__ = [
    'call_activity_task',
    'call_activity_with_retry_task',
    'call_http',
    'continue_as_new',
    'new_uuid',
    'task_all',
    'task_any',
    'should_suspend',
    'wait_for_external_event_task',
    'create_timer_task'
]



import json
import logging

logger = logging.getLogger('activity')


def main(value: int):
	logger.warn(f'ReturnSetOfInt is called with {value}')
	result = set()
	result.add(value)
	result.add(value + 1)
	result.add(value + 2)
	return result



import json
import logging

logger = logging.getLogger('activity')


def main(value: str):
    logger.warn(f'ReturnBool is called with {value}')
    if value == "1":
        return True
    else:
        return False



import json

def main(theSum: int) -> str:
    """Activity function to raise an external event to the orchestrator

    Parameters
    ----------
    theSum: int
        The sum of numbers passed to each "ParrotValue" activity function

    Returns
    -------
    str
        A string indicating the sum
    """
    return f"Well that's nice {sum(json.loads(theSum))}"


import json
from ..models.history import HistoryEventType
from ..constants import DATETIME_STRING_FORMAT
from azure.functions._durable_functions import _deserialize_custom_object


def should_suspend(partial_result) -> bool:
    """Check the state of the result to determine if the orchestration should suspend."""
    return bool(partial_result is not None
                and hasattr(partial_result, "is_completed")
                and not partial_result.is_completed)


def parse_history_event(directive_result):
    """Based on the type of event, parse the JSON.serializable portion of the event."""
    event_type = directive_result.event_type
    if event_type is None:
        raise ValueError("EventType is not found in task object")

    # We provide the ability to deserialize custom objects, because the output of this
    # will be passed directly to the orchestrator as the output of some activity
    if event_type == HistoryEventType.EVENT_RAISED:
        return json.loads(directive_result.Input, object_hook=_deserialize_custom_object)
    if event_type == HistoryEventType.SUB_ORCHESTRATION_INSTANCE_CREATED:
        return json.loads(directive_result.Result, object_hook=_deserialize_custom_object)
    if event_type == HistoryEventType.TASK_COMPLETED:
        return json.loads(directive_result.Result, object_hook=_deserialize_custom_object)
    return None


def find_event_raised(state, name):
    """Find if the event with the given event name is raised.

    Parameters
    ----------
    state : List[HistoryEvent]
        List of histories to search from
    name : str
        Name of the event to search for

    Returns
    -------
    HistoryEvent
        The raised event with the given event name that has not yet been processed.
        Returns None if no event with the given conditions was found.

    Raises
    ------
    ValueError
        Raises an error if no name was given when calling this function.
    """
    if not name:
        raise ValueError("Name cannot be empty")

    tasks = [e for e in state
             if e.event_type == HistoryEventType.EVENT_RAISED
             and e.Name == name and not e.is_processed]

    if len(tasks) == 0:
        return None

    return tasks[0]


def find_task_scheduled(state, name):
    """Locate the Scheduled Task.

    Within the state passed, search for an event that has hasn't been processed
    and has the the name provided.
    """
    if not name:
        raise ValueError("Name cannot be empty")

    tasks = [e for e in state
             if e.event_type == HistoryEventType.TASK_SCHEDULED
             and e.Name == name and not e.is_processed]

    if len(tasks) == 0:
        return None

    return tasks[0]


def find_task_completed(state, scheduled_task):
    """Locate the Completed Task.

    Within the state passed, search for an event that has hasn't been processed,
    is a completed  task type,
    and has the a scheduled id that equals the EventId of the provided scheduled task.
    """
    if scheduled_task is None:
        return None

    tasks = [e for e in state if e.event_type == HistoryEventType.TASK_COMPLETED
             and e.TaskScheduledId == scheduled_task.event_id]

    if len(tasks) == 0:
        return None

    return tasks[0]


def find_task_failed(state, scheduled_task):
    """Locate the Failed Task.

    Within the state passed, search for an event that has hasn't been processed,
    is a failed task type,
    and has the a scheduled id that equals the EventId of the provided scheduled task.
    """
    if scheduled_task is None:
        return None

    tasks = [e for e in state if e.event_type == HistoryEventType.TASK_FAILED
             and e.TaskScheduledId == scheduled_task.event_id]

    if len(tasks) == 0:
        return None

    return tasks[0]


def find_task_timer_created(state, fire_at):
    """Locate the Timer Created Task.

    Within the state passed, search for an event that has hasn't been processed,
    is a timer created task type,
    and has the an event id that is one higher then Scheduled Id of the provided
    failed task provided.
    """
    if fire_at is None:
        return None

    tasks = []
    for e in state:
        if e.event_type == HistoryEventType.TIMER_CREATED and hasattr(e, "FireAt"):
            if e.FireAt == fire_at.strftime(DATETIME_STRING_FORMAT):
                tasks.append(e)

    if len(tasks) == 0:
        return None

    return tasks[0]


def find_task_retry_timer_created(state, failed_task):
    """Locate the Timer Created Task.

    Within the state passed, search for an event that has hasn't been processed,
    is a timer created task type,
    and has the an event id that is one higher then Scheduled Id of the provided
    failed task provided.
    """
    if failed_task is None:
        return None

    tasks = [e for e in state if e.event_type == HistoryEventType.TIMER_CREATED
             and e.event_id == failed_task.TaskScheduledId + 1]

    if len(tasks) == 0:
        return None

    return tasks[0]


def find_task_retry_timer_fired(state, retry_timer_created):
    """Locate the Timer Fired Task.

    Within the state passed, search for an event that has hasn't been processed,
    is a timer fired task type,
    and has the an timer id that is equal to the EventId of the provided
    timer created task provided.
    """
    if retry_timer_created is None:
        return None

    tasks = [e for e in state if e.event_type == HistoryEventType.TIMER_FIRED
             and e.TimerId == retry_timer_created.event_id]

    if len(tasks) == 0:
        return None

    return tasks[0]


def set_processed(tasks):
    """Set the isProcessed attribute of all of the tasks to true.

    This provides the ability to not look at events that have already been processed within
    searching the history of events.
    """
    for task in tasks:
        if task is not None:
            task.is_processed = True



import json
from typing import Dict, List

from .task_utilities import find_task_scheduled, find_task_completed, find_task_failed, \
    set_processed, parse_history_event
from ..constants import HTTP_ACTION_NAME
from ..models.DurableHttpRequest import DurableHttpRequest
from ..models.TokenSource import TokenSource
from ..models.actions import CallHttpAction
from ..models.history import HistoryEvent
from ..models.Task import (
    Task)


def call_http(state: List[HistoryEvent], method: str, uri: str, content: str = None,
              headers: Dict[str, str] = None, token_source: TokenSource = None) -> Task:
    """Get task used to schedule a durable HTTP call to the specified endpoint.

    Parameters
    ----------
    state: List[HistoryEvent]
        The list of events that have been processed to determine the state of the task to be
        scheduled
    method: str
        The HTTP request method.
    uri: str
        The HTTP request uri.
    content: str
        The HTTP request content.
    headers: Dict[str, str]
        The HTTP request headers.
    token_source: TokenSource
        The source of OAuth token to add to the request.

    Returns
    -------
    Task
        The durable HTTP request to schedule.
    """
    if content and content is not isinstance(content, str):
        json_content = json.dumps(content)
    else:
        json_content = content

    request = DurableHttpRequest(method, uri, json_content, headers, token_source)

    new_action = CallHttpAction(request)

    task_scheduled = find_task_scheduled(state, HTTP_ACTION_NAME)
    task_completed = find_task_completed(state, task_scheduled)
    task_failed = find_task_failed(state, task_scheduled)
    set_processed([task_scheduled, task_completed, task_failed])

    if task_completed is not None:
        return Task(
            is_completed=True,
            is_faulted=False,
            action=new_action,
            result=parse_history_event(task_completed),
            timestamp=task_completed.timestamp,
            id_=task_completed.TaskScheduledId)

    if task_failed is not None:
        return Task(
            is_completed=True,
            is_faulted=True,
            action=new_action,
            result=task_failed.Reason,
            timestamp=task_failed.timestamp,
            id_=task_failed.TaskScheduledId,
            exc=Exception(
                f"{task_failed.Reason} \n {task_failed.Details}")
        )

    return Task(is_completed=False, is_faulted=False, action=new_action)



"""Base module for the Python Durable functions.

Exposes the different API components intended for public consumption
"""
from .orchestrator import Orchestrator
from .models.DurableOrchestrationClient import DurableOrchestrationClient
from .models.DurableOrchestrationContext import DurableOrchestrationContext
from .models.RetryOptions import RetryOptions
from .models.TokenSource import ManagedIdentityTokenSource

__all__ = [
    'Orchestrator',
    'DurableOrchestrationClient',
    'DurableOrchestrationContext',
    'ManagedIdentityTokenSource',
    'RetryOptions'
]



import json
import logging

from azure.durable_functions import DurableOrchestrationClient
import azure.functions as func


async def main(req: func.HttpRequest, starter: str) -> func.HttpResponse:
    """Activity function to raise an external event to the orchestrator
    
    Parameters
    ----------
    req: func.HttpRequest
        An HTTP Request object, it can be used to parse URL
        parameters.

    starter: str
        A JSON-formatted string describing the orchestration context

    Returns
    -------
    func.HttpResponse
        HTTP response object whose body indicates which event
        was raised
    """

    logging.info("Recevied http request to check startus {}".format(starter))
    client = DurableOrchestrationClient(starter)
    instance_id = req.params.get("instance_id")
    logging.info("Will check on instance id: {}".format(instance_id))
    
    event_name = req.params.get("event_name")
    logging.info("Will check on event: {}".format(event_name))

    await client.raise_event(instance_id, event_name, True)
    return func.HttpResponse(f'"{event_name}" event is sent')



from typing import Any

from ..models.Task import (
    Task)
from ..models.actions.ContinueAsNewAction import ContinueAsNewAction


def continue_as_new(
        input_: Any = None) -> Task:
    """Create a new continue as new action.

    Parameters
    ----------
    input_: Any
        The JSON-serializable input to pass to the activity function.

    Returns
    -------
    Task
        A Durable Task that causes the orchestrator reset and start as a new orchestration.
    """
    new_action = ContinueAsNewAction(input_)

    return Task(is_completed=False, is_faulted=False, action=new_action)



from datetime import datetime

from .actions import Action


class Task:
    """Represents some pending action.

    Similar to a native JavaScript promise in
    that it acts as a placeholder for outstanding asynchronous work, but has
    a synchronous implementation and is specific to Durable Functions.

    Tasks are only returned to an orchestration function when a
    [[DurableOrchestrationContext]] operation is not called with `yield`. They
    are useful for parallelization and timeout operations in conjunction with
    Task.all and Task.any.
    """

    def __init__(self, is_completed, is_faulted, action,
                 result=None, timestamp=None, id_=None, exc=None):
        self._is_completed: bool = is_completed
        self._is_faulted: bool = is_faulted
        self._action: Action = action
        self._result = result
        self._timestamp: datetime = timestamp
        self._id = id_
        self._exception = exc

    @property
    def is_completed(self) -> bool:
        """Get indicator whether the task has completed.

        Note that completion is not equivalent to success.
        """
        return self._is_completed

    @property
    def is_faulted(self) -> bool:
        """Get indicator whether the task faulted in some way due to error."""
        return self._is_faulted

    @property
    def action(self) -> Action:
        """Get the scheduled action represented by the task.

        _Internal use only._
        """
        return self._action

    @property
    def result(self) -> object:
        """Get the result of the task, if completed. Otherwise `None`."""
        return self._result

    @property
    def timestamp(self) -> datetime:
        """Get the timestamp of the task."""
        return self._timestamp

    @property
    def id(self):
        """Get the ID number of the task.

        _Internal use only._
        """
        return self._id

    @property
    def exception(self):
        """Get the error thrown when attempting to perform the task's action.

        If the Task has not yet completed or has completed successfully, `None`
        """
        return self._exception



from ..models.Task import Task


class TimerTask(Task):
    """Represents a pending timer.

    All pending timers must be completed or canceled for an orchestration to complete.

    Example: Cancel a timer
    ```
    timeout_task = context.df.create_timer(expiration_date)
    if not timeout_task.is_completed():
        timeout_task.cancel()
    ```
    """

    def __init__(self, action, is_completed, timestamp, id_):
        self._action = action
        self._is_completed = is_completed
        self._timestamp = timestamp
        self._id = id_

        super().__init__(self._is_completed, False,
                         self._action, None, self._timestamp, self._id, None)

    def is_cancelled(self) -> bool:
        """Check of a timer is cancelled.

        Returns
        -------
        bool
            Returns whether a timer has been cancelled or not
        """
        return self._action.is_cancelled

    def cancel(self):
        """Cancel a timer.

        Raises
        ------
        ValueError
            Raises an error if the task is already completed and an attempt is made to cancel it
        """
        if not self._is_completed:
            self._action.is_cancelled = True
        else:
            raise ValueError("Cannot cancel a completed task.")



import json
import logging
from typing import List
import azure.durable_functions as df
import azure.functions as func


def orchestrator_function(context: df.DurableOrchestrationContext) -> List[str]:

    """This function provides the core function chaining orchestration logic

    Parameters
    ----------
    context: DurableOrchestrationContext
        This context has the past history and the durable orchestration API

    Returns
    -------
    output: List[str]
        Returns an array of result by the activity functions.

    Yields
    -------
    call_activity: str
        Yields, depending on the `json_rule`, to wait on either all
        tasks to complete, or until one of the tasks completes.
    """


    logging.debug("Creating the orchestrator function")

    json_rule = {
        "condition": {
            "wait_events": ["A","B"],
            "logic": "and"
        },
        "satisfied":[
            {
                "activity_func_name": "SuccessActions",
                "args": {
                    "name": "Tokyo"
                }
            }
        ]
    }

    tasks = []
    for event in json_rule["condition"]["wait_events"]:
        tasks.append(context.wait_for_external_event(event))
        logging.debug("Added event {} to list of tasks".format(event))

    if json_rule["condition"]["logic"] == 'and':
        logging.info("A logical <and> rule was found")
        yield context.task_all(tasks)
    elif json_rule["condition"]["logic"] == 'or':
        logging.info("A logical <or> rule was found")
        yield context.task_any(tasks)

    output = []
    for action in json_rule["satisfied"]:
        result = yield context.call_activity(action["activity_func_name"], json.dumps(action["args"]))
        output.append(result)

    return output


main = df.Orchestrator.create(orchestrator_function)



import json
from datetime import datetime

from dateutil.tz import tzutc

from azure.durable_functions.models.Task import Task
from azure.durable_functions.models.actions.WaitForExternalEventAction import \
    WaitForExternalEventAction
from azure.durable_functions.tasks.wait_for_external_event import wait_for_external_event_task
from tests.test_utils.ContextBuilder import ContextBuilder
from .tasks_test_utils import assert_tasks_equal


def test_event_not_raised_return_incompleted_task():
    context_builder = ContextBuilder('test_simple_function')
    expected_action = WaitForExternalEventAction("A")

    returned_task = wait_for_external_event_task(context_builder.history_events, "A")
    expected_task = Task(is_completed=False, is_faulted=False, action=expected_action)

    assert_tasks_equal(expected_task, returned_task)


def test_event_raised_return_completed_task():
    timestamp = datetime.now()
    json_input = '{"test":"somecontent"}'
    expected_action = WaitForExternalEventAction("A")
    context_builder = ContextBuilder('test_simple_function')
    context_builder.add_event_raised_event(name="A", input_=json_input, timestamp=timestamp, id_=1)

    returned_task = wait_for_external_event_task(context_builder.history_events, "A")
    expected_task = Task(
        is_completed=True,
        is_faulted=False,
        action=expected_action,
        result=json.loads(json_input),
        timestamp=timestamp.replace(tzinfo=tzutc()),
        id_=1)

    assert_tasks_equal(expected_task, returned_task)



from uuid import uuid5, NAMESPACE_OID

from azure.durable_functions.models import DurableOrchestrationContext
from azure.durable_functions.constants import DATETIME_STRING_FORMAT

URL_NAMESPACE: str = "9e952958-5e33-4daf-827f-2fa12937b875"


def _create_deterministic_uuid(namespace_value: str, name: str) -> str:
    namespace_uuid = uuid5(NAMESPACE_OID, namespace_value)
    return str(uuid5(namespace_uuid, name))


def new_uuid(context: DurableOrchestrationContext) -> str:
    """Create a new UUID that is safe for replay within an orchestration or operation.

    The default implementation of this method creates a name-based UUID
    using the algorithm from RFC 4122 §4.3. The name input used to generate
    this value is a combination of the orchestration instance ID and an
    internally managed sequence number.

    Parameters
    ----------
    context : DurableOrchestrationContext
        Provides reference to the instance id, current_utc_datetime and a new_uuid_counter
        attribute that is combined together to form that name that is used for the V5 UUID.

    Returns
    -------
    str
        New UUID that is safe for replay within an orchestration or operation.
    """
    uuid_name_value = \
        f"{context.instance_id}" \
        f"_{context.current_utc_datetime.strftime(DATETIME_STRING_FORMAT)}" \
        f"_{context._new_uuid_counter}"
    context._new_uuid_counter += 1
    return _create_deterministic_uuid(URL_NAMESPACE, uuid_name_value)



import json
import datetime
from typing import List, Any, Dict

from . import (RetryOptions, TaskSet)
from .FunctionContext import FunctionContext
from .history import HistoryEvent, HistoryEventType
from .actions import Action
from ..models.Task import Task
from ..models.TokenSource import TokenSource
from ..tasks import call_activity_task, task_all, task_any, call_activity_with_retry_task, \
    wait_for_external_event_task, continue_as_new, new_uuid, call_http, create_timer_task
from azure.functions._durable_functions import _deserialize_custom_object


class DurableOrchestrationContext:
    """Context of the durable orchestration execution.

    Parameter data for orchestration bindings that can be used to schedule
    function-based activities.
    """

    # parameter names are as defined by JSON schema and do not conform to PEP8 naming conventions
    def __init__(self,
                 history: List[Dict[Any, Any]], instanceId: str, isReplaying: bool,
                 parentInstanceId: str, input: Any = None, **kwargs):
        self._histories: List[HistoryEvent] = [HistoryEvent(**he) for he in history]
        self._instance_id: str = instanceId
        self._is_replaying: bool = isReplaying
        self._parent_instance_id: str = parentInstanceId
        self._custom_status: Any = None
        self._new_uuid_counter: int = 0
        self.call_activity = lambda n, i=None: call_activity_task(
            state=self.histories,
            name=n,
            input_=i)
        self.call_activity_with_retry = \
            lambda n, o, i=None: call_activity_with_retry_task(
                state=self.histories,
                retry_options=o,
                name=n,
                input_=i)
        self.call_http = lambda method, uri, content=None, headers=None, token_source=None: \
            call_http(
                state=self.histories, method=method, uri=uri, content=content, headers=headers,
                token_source=token_source)
        self.wait_for_external_event = lambda n: wait_for_external_event_task(
            state=self.histories,
            name=n)
        self.new_uuid = lambda: new_uuid(context=self)
        self.continue_as_new = lambda i: continue_as_new(input_=i)
        self.task_any = lambda t: task_any(tasks=t)
        self.task_all = lambda t: task_all(tasks=t)
        self.create_timer = lambda d: create_timer_task(state=self.histories, fire_at=d)
        self.decision_started_event: HistoryEvent = \
            [e_ for e_ in self.histories
             if e_.event_type == HistoryEventType.ORCHESTRATOR_STARTED][0]
        self._current_utc_datetime = \
            self.decision_started_event.timestamp
        self._new_uuid_counter = 0
        self.actions: List[List[Action]] = []
        self._function_context: FunctionContext = FunctionContext(**kwargs)

        # make _input always a string
        # (consistent with Python Functions generic trigger/input bindings)
        if (isinstance(input, Dict)):
            input = json.dumps(input)
        self._input: str = input

    @classmethod
    def from_json(cls, json_string: str):
        """Convert the value passed into a new instance of the class.

        Parameters
        ----------
        json_string: str
            Context passed a JSON serializable value to be converted into an instance of the class

        Returns
        -------
        DurableOrchestrationContext
            New instance of the durable orchestration context class
        """
        # We should consider parsing the `Input` field here as well,
        # intead of doing so lazily when `get_input` is called.
        json_dict = json.loads(json_string)
        return cls(**json_dict)

    def call_activity(self, name: str, input_=None) -> Task:
        """Schedule an activity for execution.

        Parameters
        ----------
        name: str
            The name of the activity function to call.
        input_:
            The JSON-serializable input to pass to the activity function.

        Returns
        -------
        Task
            A Durable Task that completes when the called activity function completes or fails.
        """
        raise NotImplementedError("This is a placeholder.")

    def call_activity_with_retry(self,
                                 name: str, retry_options: RetryOptions,
                                 input_=None) -> Task:
        """Schedule an activity for execution with retry options.

        Parameters
        ----------
        name: str
            The name of the activity function to call.
        retry_options: RetryOptions
            The retry options for the activity function.
        input_:
            The JSON-serializable input to pass to the activity function.

        Returns
        -------
        Task
            A Durable Task that completes when the called activity function completes or
            fails completely.
        """
        raise NotImplementedError("This is a placeholder.")

    def call_http(self, method: str, uri: str, content: str = None,
                  headers: Dict[str, str] = None, token_source: TokenSource = None) -> Task:
        """Schedule a durable HTTP call to the specified endpoint.

        Parameters
        ----------
        method: str
            The HTTP request method.
        uri: str
            The HTTP request uri.
        content: str
            The HTTP request content.
        headers: Dict[str, str]
            The HTTP request headers.
        token_source: TokenSource
            The source of OAuth token to add to the request.

        Returns
        -------
        Task
            The durable HTTP request to schedule.
        """
        raise NotImplementedError("This is a placeholder.")

    def call_sub_orchestrator(self,
                              name: str, input_=None,
                              instance_id: str = None) -> Task:
        """Schedule an orchestration function named `name` for execution.

        Parameters
        ----------
        name: str
            The name of the orchestrator function to call.
        input_:
            The JSON-serializable input to pass to the orchestrator function.
        instance_id: str
            A unique ID to use for the sub-orchestration instance. If `instanceId` is not
            specified, the extension will generate an id in the format `<calling orchestrator
            instance ID>:<#>`
        """
        raise NotImplementedError("This is a placeholder.")

    def get_input(self) -> str:
        """Get the orchestration input."""
        return None if self._input is None else json.loads(self._input,
                                                           object_hook=_deserialize_custom_object)

    def new_uuid(self) -> str:
        """Create a new UUID that is safe for replay within an orchestration or operation.

        The default implementation of this method creates a name-based UUID
        using the algorithm from RFC 4122 §4.3. The name input used to generate
        this value is a combination of the orchestration instance ID and an
        internally managed sequence number.

        Returns
        -------
        str
            New UUID that is safe for replay within an orchestration or operation.
        """
        raise NotImplementedError("This is a placeholder.")

    def task_all(self, activities: List[Task]) -> TaskSet:
        """Schedule the execution of all activities.

        Similar to Promise.all. When called with `yield` or `return`, returns an
        array containing the results of all [[Task]]s passed to it. It returns
        when all of the [[Task]] instances have completed.

        Throws an exception if any of the activities fails
        Parameters
        ----------
        activities: List[Task]
            List of activities to schedule

        Returns
        -------
        TaskSet
            The results of all activities.
        """
        raise NotImplementedError("This is a placeholder.")

    def task_any(self, activities: List[Task]) -> TaskSet:
        """Schedule the execution of all activities.

        Similar to Promise.race. When called with `yield` or `return`, returns
        the first [[Task]] instance to complete.

        Throws an exception if all of the activities fail

        Parameters
        ----------
        activities: List[Task]
            List of activities to schedule

        Returns
        -------
        TaskSet
            The first [[Task]] instance to complete.
        """
        raise NotImplementedError("This is a placeholder.")

    def set_custom_status(self, status: Any):
        """Set the customized orchestration status for your orchestrator function.

        This status is also returned by the orchestration client through the get_status API

        Parameters
        ----------
        status : str
            Customized status provided by the orchestrator
        """
        self._custom_status = status

    @property
    def custom_status(self):
        """Get customized status of current orchestration."""
        return self._custom_status

    @property
    def histories(self):
        """Get running history of tasks that have been scheduled."""
        return self._histories

    @property
    def instance_id(self) -> str:
        """Get the ID of the current orchestration instance.

        The instance ID is generated and fixed when the orchestrator function
        is scheduled. It can be either auto-generated, in which case it is
        formatted as a GUID, or it can be user-specified with any format.

        Returns
        -------
        str
            The ID of the current orchestration instance.
        """
        return self._instance_id

    @property
    def is_replaying(self) -> bool:
        """Get the value indicating orchestration replaying itself.

        This property is useful when there is logic that needs to run only when
        the orchestrator function is _not_ replaying. For example, certain
        types of application logging may become too noisy when duplicated as
        part of orchestrator function replay. The orchestrator code could check
        to see whether the function is being replayed and then issue the log
        statements when this value is `false`.

        Returns
        -------
        bool
            Value indicating whether the orchestrator function is currently replaying.
        """
        return self._is_replaying

    @property
    def parent_instance_id(self) -> str:
        """Get the ID of the parent orchestration.

        The parent instance ID is generated and fixed when the parent
        orchestrator function is scheduled. It can be either auto-generated, in
        which case it is formatted as a GUID, or it can be user-specified with
        any format.

        Returns
        -------
        str
            ID of the parent orchestration of the current sub-orchestration instance
        """
        return self._parent_instance_id

    @property
    def current_utc_datetime(self) -> datetime:
        """Get the current date/time.

        This date/time value is derived from the orchestration history. It
        always returns the same value at specific points in the orchestrator
        function code, making it deterministic and safe for replay.

        Returns
        -------
        datetime
            The current date/time in a way that is safe for use by orchestrator functions
        """
        return self._current_utc_datetime

    @current_utc_datetime.setter
    def current_utc_datetime(self, value: datetime):
        self._current_utc_datetime = value

    @property
    def function_context(self) -> FunctionContext:
        """Get the function level attributes not used by durable orchestrator.

        Returns
        -------
        FunctionContext
            Object containing function level attributes not used by durable orchestrator.
        """
        return self._function_context



import logging

def main(key: str) -> str:
    """Activity function performing a specific step in the chain
    
    Parameters
    ----------
    key : str
        key to a dictionary that advances the chain
    
    Returns
    -------
    [int]
        value in the dictionary as a result
    """

    logging.warning(f"Activity Triggered: {key}")
    switch_statement = {}
    switch_statement["One"] = "Two"
    switch_statement["Two"] = "Three"
    try:
        if switch_statement[key]:
            return f'{switch_statement[key]}'
    except KeyError as e:
        return f'One'

    



import json
import logging

logger = logging.getLogger('activity')


def main(value: str):
    logger.warn(f'ReturnDictOfString is called with {value}')
    result = {}
    result[value] = value
    return result



import json

def main(value: str) -> str:
    """Activity function to generate a range of numbers

    Parameters
    ----------
    value: str
        The exclusive upper-bound of the generated range of numbers

    Returns
    -------
    str
        A JSON-formatted string representing the range of values:
        [0-(value -1)]
    """
    activity_values = [*range(int(value))]
    return json.dumps(activity_values)



from typing import List
from ..models.Task import Task
from ..models.TaskSet import TaskSet


def task_all(tasks: List[Task]):
    """Determine the state of scheduling the activities for execution with retry options.

    Parameters
    ----------
    tasks: List[Task]
        The tasks to evaluate their current state.

    Returns
    -------
    TaskSet
        A Durable Task Set that reports the state of running all of the tasks within it.
    """
    all_actions = []
    results = []
    is_completed = True
    complete_time = None
    faulted = []
    for task in tasks:
        if isinstance(task, TaskSet):
            for action in task.actions:
                all_actions.append(action)
        else:
            all_actions.append(task.action)
        results.append(task.result)

        if task.is_faulted:
            faulted.append(task.exception)

        if not task.is_completed:
            is_completed = False
        else:
            complete_time = task.timestamp if complete_time is None \
                else max([task.timestamp, complete_time])

    if len(faulted) > 0:
        return TaskSet(is_completed, all_actions, results, is_faulted=True, exception=faulted[0])
    if is_completed:
        return TaskSet(is_completed, all_actions, results, False, complete_time)
    else:
        return TaskSet(is_completed, all_actions, None)



import json
from typing import List
import azure.functions as func
import azure.durable_functions as df


def _get_classify_images_tasks(config: dict, image_list: List[str], context: df.DurableOrchestrationContext):
    """Get list of tasks that breaks down the execution of the predications.

    will create a list of tasks to perform that is split evenly across the
    different instances

    Parameters
    ----------
    config: dict
        Describes how the tasks will be split
    image_list: List[str]
        The list of images to classify
    context: df.DurableOrchestrationContext
        The Durable context to call the activities from

    Returns
    -------
    tasks: List
        List of tasks to perform
    """
    image_count_per_instance = int(
        config['number_of_images']/config['instances'])

    tasks = []

    start = 0
    increment = image_count_per_instance

    for _ in range(config['instances']):
        instance_images = image_list[start:increment]
        tasks.append(
            context.call_activity("ClassifyImage",
                                  instance_images))
        start += image_count_per_instance
        increment += image_count_per_instance

    return tasks


def orchestrator_function(context: df.DurableOrchestrationContext):
    """Get the generator that will need to be orchestrated by durable functions.

    This function will get a list of images to do a prediction of, fan out the
    prediction tasks then summarize the results

    Parameters
    ----------
    context: df.DurableOrchestrationContext
        The Durable context to perform the activities with

    Returns
    -------
    summary
        A summary of the prediction results

    Yields
    -------
        Tasks that need to be performed by the Durable orchestrator
    """
    config = {
        "instances": 5,  # The number of instances to fan out the prediction tasks
        "number_of_images": 15  # The number of images to predict
    }

    # Get the images that need to predicted
    image_data = yield context.call_activity("GetImageUrls", config['number_of_images'])
    image_list = json.loads(image_data)

    # break the images done into different tasks to be fan out with
    tasks = _get_classify_images_tasks(config, image_list, context)
    predictions = yield context.task_all(tasks)

    # combine the results of the predictions into a single list
    combined = []
    for tr in predictions:
        prediction = json.loads(tr)
        combined.extend(prediction)

    # summarize the results
    summary = yield context.call_activity("ShowMeTheResults", combined)
    return summary


main = df.Orchestrator.create(orchestrator_function)



from datetime import datetime, timedelta
import json
from .predict import predict_image_from_url


def main(value: str) -> str:
    """Classify the list of images based on whether they are a dog or cat
    
    Parameters
    ----------
    value: str
        List of image URLs to predict
    
    Returns
    -------
    str
        JSON-formatted string of the prediction results
    """
    images = json.loads(value)

    prediction_results = []
    for image_url in images:
        results = predict_image_from_url(image_url)
        if results is not None:
            prediction_results.append({
                'tag': results['predictedTagName'],
                'url': image_url
            })
        else:
            prediction_results.append({
                'tag': 'error',
                'url': image_url
            })

    return json.dumps(prediction_results)



from typing import List, Any

from ..models.Task import (
    Task)
from ..models.actions.CallActivityAction import CallActivityAction
from ..models.history import HistoryEvent
from .task_utilities import find_task_completed, find_task_failed, \
    find_task_scheduled, set_processed, parse_history_event


def call_activity_task(
        state: List[HistoryEvent],
        name: str,
        input_: Any = None) -> Task:
    """Determine the state of Scheduling an activity for execution.

    Parameters
    ----------
    state: List[HistoryEvent]
        The list of history events to search to determine the current state of the activity.
    name: str
        The name of the activity function to schedule.
    input_: Any
        The JSON-serializable input to pass to the activity function.

    Returns
    -------
    Task
        A Durable Task that completes when the called activity function completes or fails.
    """
    new_action = CallActivityAction(name, input_)

    task_scheduled = find_task_scheduled(state, name)
    task_completed = find_task_completed(state, task_scheduled)
    task_failed = find_task_failed(state, task_scheduled)
    set_processed([task_scheduled, task_completed, task_failed])

    if task_completed is not None:
        return Task(
            is_completed=True,
            is_faulted=False,
            action=new_action,
            result=parse_history_event(task_completed),
            timestamp=task_completed.timestamp,
            id_=task_completed.TaskScheduledId)

    if task_failed is not None:
        return Task(
            is_completed=True,
            is_faulted=True,
            action=new_action,
            result=task_failed.Reason,
            timestamp=task_failed.timestamp,
            id_=task_failed.TaskScheduledId,
            exc=Exception(
                f"{task_failed.Reason} \n {task_failed.Details}")
        )

    return Task(is_completed=False, is_faulted=False, action=new_action)



import json
from typing import Callable, Iterator, Any, Dict
from jsonschema import validate

from azure.durable_functions.models import DurableOrchestrationContext
from azure.durable_functions.orchestrator import Orchestrator
from .schemas.OrchetrationStateSchema import schema


def assert_orchestration_state_equals(expected, result):
    assert_attribute_equal(expected, result, "isDone")
    assert_actions_are_equal(expected, result)
    assert_attribute_equal(expected, result, "output")
    assert_attribute_equal(expected, result, "error")
    assert_attribute_equal(expected, result, "customStatus")


def assert_attribute_equal(expected, result, attribute):
    if attribute in expected:
        assert result.get(attribute) == expected.get(attribute)
    else:
        assert attribute not in result


def assert_actions_are_equal(expected, result):
    expected_actions = expected.get("actions")
    result_actions = result.get("actions")
    assert len(expected_actions) == len(result_actions)
    for index in range(len(expected_actions)):
        assert len(expected_actions[index]) == len(result_actions[index])
        for action_index in range(len(expected_actions[index])):
            expected_action = expected_actions[index][action_index]
            result_action = result_actions[index][action_index]
            assert_action_is_equal(expected_action, result_action)


def assert_action_is_equal(expected_action, result_action):
    assert_attribute_equal(expected_action, result_action, "functionName")
    assert_attribute_equal(expected_action, result_action, "input")
    assert_attribute_equal(expected_action, result_action, "actionType")


def get_orchestration_state_result(
        context_builder,
        activity_func: Callable[[DurableOrchestrationContext], Iterator[Any]]):
    context_as_string = context_builder.to_json_string()
    orchestrator = Orchestrator(activity_func)
    result_of_handle = orchestrator.handle(
        DurableOrchestrationContext.from_json(context_as_string))
    result = json.loads(result_of_handle)
    return result


def assert_valid_schema(orchestration_state):
    validation_results = validate(instance=orchestration_state, schema=schema)
    assert validation_results is None


def assert_dict_are_equal(expected: Dict[Any, Any], result: Dict[Any, Any]):
    assert len(expected.keys()) == len(result.keys())
    for key in expected.keys():
        assert expected[key] == result[key]
    for key in result.keys():
        assert result[key] == expected[key]



import json

import azure.functions as func
import azure.durable_functions as df

def orchestrator_function(context: df.DurableOrchestrationContext):
    """This function provides the core fan-out-fan-in orchestration logic

    Parameters
    ----------
    context: DurableOrchestrationContext
        This context has the past history and the durable orchestration API

    Returns
    -------
    message
        Returns the result of the "ShowMeTheSum" activity function.

    Yields
    -------
    call_activity: str
        Yields, depending on the `json_rule`, to wait on either all
        tasks to complete, or until one of the tasks completes.
    """

    activity_count = yield context.call_activity("GetActivityCount", 5)
    activity_list = json.loads(activity_count)

    tasks = [context.call_activity("ParrotValue", i) for i in activity_list]

    tasks_result = yield context.task_all(tasks)
    values = [int(t) for t in tasks_result]
    message = yield context.call_activity("ShowMeTheSum", values)

    return message


main = df.Orchestrator.create(orchestrator_function)



from typing import Dict
import json
import logging

logger = logging.getLogger('activity')


def main(value: str) -> Dict[str, str]:
    logger.warn(f'ReturnDictOfStringWithAnnotation is called with {value}')
    result = {}
    result[value] = value
    return result



from .orchestrator_test_utils \
    import assert_orchestration_state_equals, get_orchestration_state_result, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.actions.CallActivityAction \
    import CallActivityAction
from tests.test_utils.testClasses import SerializableClass


def generator_function(context):
    outputs = []

    task1 = yield context.call_activity("Hello", "Tokyo")
    task2 = yield context.call_activity("Hello", "Seattle")
    task3 = yield context.call_activity("Hello", "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs


def generator_function_with_serialization(context):
    """Ochestrator to test sequential activity calls with a serializable input arguments."""
    outputs = []

    task1 = yield context.call_activity("Hello", SerializableClass("Tokyo"))
    task2 = yield context.call_activity("Hello", SerializableClass("Seattle"))
    task3 = yield context.call_activity("Hello", SerializableClass("London"))

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs


def base_expected_state(output=None) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output)


def add_hello_action(state: OrchestratorState, input_: str):
    action = CallActivityAction(function_name='Hello', input_=input_)
    state.actions.append([action])


def add_hello_completed_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=id_, result=result)


def add_hello_failed_events(
        context_builder: ContextBuilder, id_: int, reason: str, details: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_failed_event(
        id_=id_, reason=reason, details=details)


def test_initial_orchestration_state():
    context_builder = ContextBuilder('test_simple_function')

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_state():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(
        context_builder, 0, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected_state._error = f'{failed_reason} \n {failed_details}'
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_and_seattle_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_completed_events(context_builder, 1, "\"Hello Seattle!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    add_hello_action(expected_state, 'London')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_and_seattle_and_london_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_completed_events(context_builder, 1, "\"Hello Seattle!\"")
    add_hello_completed_events(context_builder, 2, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(
        ['Hello Tokyo!', 'Hello Seattle!', 'Hello London!'])
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    add_hello_action(expected_state, 'London')
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_and_seattle_and_london_with_serialization_state():
    """Tests the sequential function pattern with custom object serialization.
    
    This simple test validates that a sequential function pattern returns
    the expected state when the input to activities is a user-provided
    serializable class.
    """
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_completed_events(context_builder, 1, "\"Hello Seattle!\"")
    add_hello_completed_events(context_builder, 2, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function_with_serialization)

    expected_state = base_expected_state(
        ['Hello Tokyo!', 'Hello Seattle!', 'Hello London!'])
    add_hello_action(expected_state, SerializableClass("Tokyo"))
    add_hello_action(expected_state, SerializableClass("Seattle"))
    add_hello_action(expected_state, SerializableClass("London"))
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)



import logging
import json

def main(args: str) -> str:
    """Activity function to raise an external event to the orchestrator

    Parameters
    ----------
    req: func.HttpRequest
        An HTTP Request object, it can be used to parse URL
        parameters.

    Returns
    -------
    str
        A 'Hello-string' to the argument passed in via args
    """
    logging.info(f"Activity Triggered: SuccessActions")

    args= json.loads(args)
    logging.info("Activity arguments: {}".format(args))
    return f'Hello {args["name"]}'



import json

def main(value: str) -> str:
    """Get a summary of the results of the predictions.

    Parameters
    ----------
    value: str
        List-formatted string of the predictions

    Returns
    -------
    str
        JSON-formatted string representing the summary of predictions
    """
    results = json.loads(value)
    analysis = {}
    analysis['images_processed'] = len(results)
    dogs = [d for d in results if d['tag'] == 'dog']
    cats = [c for c in results if c['tag'] == 'cat']
    error = [e for e in results if e['tag'] == 'error']
    analysis['number_of_dogs'] = len(dogs)
    analysis['number_of_cats'] = len(cats)
    analysis['number_failed'] = len(error)
    return json.dumps(analysis)



from datetime import datetime, date
import json
from azure.durable_functions.models import Task, TaskSet
from azure.durable_functions.tasks import task_any
from azure.durable_functions.tasks.wait_for_external_event import wait_for_external_event_task
from azure.durable_functions.models.actions.WaitForExternalEventAction import WaitForExternalEventAction
from azure.durable_functions.constants import DATETIME_STRING_FORMAT
from tests.test_utils.ContextBuilder import ContextBuilder
from .tasks_test_utils import assert_taskset_equal


from tests.orchestrator.orchestrator_test_utils \
    import assert_orchestration_state_equals, get_orchestration_state_result
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from tests.orchestrator.test_sequential_orchestrator import base_expected_state,\
 add_hello_action, add_hello_failed_events

def test_has_completed_task():
    all_actions = [WaitForExternalEventAction("C"), WaitForExternalEventAction("A"), WaitForExternalEventAction("B")]
    task1 = Task(is_completed=False, is_faulted=False, action=all_actions[0], timestamp=date(2000,1,1))
    task2 = Task(is_completed=True, is_faulted=False, action=all_actions[1],timestamp=date(2000,2,1))
    task3 = Task(is_completed=True, is_faulted=False, action=all_actions[2],timestamp=date(2000,1,1))

    tasks = [task1, task2, task3]
    returned_taskset = task_any(tasks)
    expected_taskset = TaskSet(is_completed=True, actions=all_actions, result=task3, timestamp=date(2000,1,1))

    assert_taskset_equal(expected_taskset, returned_taskset)

def test_has_no_completed_task():
    all_actions = [WaitForExternalEventAction("C"), WaitForExternalEventAction("A"), WaitForExternalEventAction("B")]
    task1 = Task(is_completed=False, is_faulted=False, action=all_actions[0], timestamp=date(2000,1,1))
    task2 = Task(is_completed=False, is_faulted=False, action=all_actions[1],timestamp=date(2000,2,1))
    task3 = Task(is_completed=False, is_faulted=False, action=all_actions[2],timestamp=date(2000,1,1))

    tasks = [task1, task2, task3]
    returned_taskset = task_any(tasks)
    expected_taskset = TaskSet(is_completed=False, actions=all_actions, result=None)

    assert_taskset_equal(expected_taskset, returned_taskset)

def test_all_faulted_task_should_fail():
    all_actions = [WaitForExternalEventAction("C"), WaitForExternalEventAction("A"), WaitForExternalEventAction("B")]
    task1 = Task(is_completed=False, is_faulted=True, action=all_actions[0], timestamp=date(2000,1,1), exc=Exception("test failure"))
    task2 = Task(is_completed=False, is_faulted=True, action=all_actions[1], timestamp=date(2000,2,1), exc=Exception("test failure"))
    task3 = Task(is_completed=False, is_faulted=True, action=all_actions[2], timestamp=date(2000,1,1), exc=Exception("test failure"))

    tasks = [task1, task2, task3]
    returned_taskset = task_any(tasks)
    error_messages = [Exception("test failure") for _ in range(3)]
    expected_exception = Exception(f"All tasks have failed, errors messages in all tasks:{error_messages}")
    expected_taskset = TaskSet(is_completed=True, actions=all_actions, result=None, is_faulted=True, exception=expected_exception)
    assert_taskset_equal(expected_taskset, returned_taskset)

def test_one_faulted_task_should_still_proceed():
    all_actions = [WaitForExternalEventAction("C"), WaitForExternalEventAction("A"), WaitForExternalEventAction("B")]
    task1 = Task(is_completed=False, is_faulted=True, action=all_actions[0], timestamp=date(2000,1,1))
    task2 = Task(is_completed=False, is_faulted=False, action=all_actions[1],timestamp=date(2000,2,1))
    task3 = Task(is_completed=False, is_faulted=False, action=all_actions[2],timestamp=date(2000,1,1))

    tasks = [task1, task2, task3]
    returned_taskset = task_any(tasks)
    expected_taskset = TaskSet(is_completed=False, actions=all_actions, result=None)

    assert_taskset_equal(expected_taskset, returned_taskset)

def test_taskset_and_tasks_as_args():
    all_actions = [WaitForExternalEventAction("C"), WaitForExternalEventAction("A"), WaitForExternalEventAction("B")]
    task1 = Task(is_completed=False, is_faulted=True, action=all_actions[0], timestamp=date(2000,1,1))
    task2 = TaskSet(is_completed=True, is_faulted=False, actions=[all_actions[1], all_actions[2]], \
            result=[None, None], timestamp=date(2000,1,1))

    tasks = [task1, task2]
    returned_taskset = task_any(tasks)
    expected_taskset = TaskSet(is_completed=True, actions=all_actions, result=task2, timestamp=date(2000,1,1))

    assert_taskset_equal(expected_taskset, returned_taskset)



from typing import Dict, Any

from azure.durable_functions.models import TokenSource
from azure.durable_functions.models.utils.json_utils import add_attrib, add_json_attrib


class DurableHttpRequest:
    """Data structure representing a durable HTTP request."""

    def __init__(self, method: str, uri: str, content: str = None, headers: Dict[str, str] = None,
                 token_source: TokenSource = None):
        self._method: str = method
        self._uri: str = uri
        self._content: str = content
        self._headers: Dict[str, str] = headers
        self._token_source: TokenSource = token_source

    @property
    def method(self) -> str:
        """Get the HTTP request method."""
        return self._method

    @property
    def uri(self) -> str:
        """Get the HTTP request uri."""
        return self._uri

    @property
    def content(self) -> str:
        """Get the HTTP request content."""
        return self._content

    @property
    def headers(self) -> Dict[str, str]:
        """Get the HTTP request headers."""
        return self._headers

    @property
    def token_source(self) -> TokenSource:
        """Get the source of OAuth token to add to the request."""
        return self._token_source

    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict = {}
        add_attrib(json_dict, self, 'method')
        add_attrib(json_dict, self, 'uri')
        add_attrib(json_dict, self, 'content')
        add_attrib(json_dict, self, 'headers')
        add_json_attrib(json_dict, self, 'token_source', 'tokenSource')
        return json_dict



"""Durable Orchestrator.

Responsible for orchestrating the execution of the user defined generator
function.
"""
from typing import Callable, Iterator, Any

from .models import (
    DurableOrchestrationContext,
    Task,
    TaskSet,
    OrchestratorState)
from .models.history import HistoryEventType
from .tasks import should_suspend

import azure.functions as func


class Orchestrator:
    """Durable Orchestration Class.

    Responsible for orchestrating the execution of the user defined generator
    function.
    """

    def __init__(self,
                 activity_func: Callable[[DurableOrchestrationContext], Iterator[Any]]):
        """Create a new orchestrator for the user defined generator.

        Responsible for orchestrating the execution of the user defined
        generator function.
        :param activity_func: Generator function to orchestrate.
        """
        self.fn: Callable[[DurableOrchestrationContext], Iterator[Any]] = activity_func

    def handle(self, context: DurableOrchestrationContext):
        """Handle the orchestration of the user defined generator function.

        Called each time the durable extension executes an activity and needs
        the client to handle the result.

        :param context: the context of what has been executed by
        the durable extension.
        :return: the resulting orchestration state, with instructions back to
        the durable extension.
        """
        self.durable_context = context
        self.generator = None
        suspended = False

        fn_output = self.fn(self.durable_context)
        # If `fn_output` is not an Iterator, then the orchestrator
        # function does not make use of its context parameter. If so,
        # `fn_output` is the return value instead of a generator
        if isinstance(fn_output, Iterator):
            self.generator = fn_output

        else:
            orchestration_state = OrchestratorState(
                is_done=True,
                output=fn_output,
                actions=self.durable_context.actions,
                custom_status=self.durable_context.custom_status)
            return orchestration_state.to_json_string()
        try:
            generation_state = self._generate_next(None)

            while not suspended:
                self._add_to_actions(generation_state)

                if should_suspend(generation_state):
                    orchestration_state = OrchestratorState(
                        is_done=False,
                        output=None,
                        actions=self.durable_context.actions,
                        custom_status=self.durable_context.custom_status)
                    suspended = True
                    continue

                if (isinstance(generation_state, Task)
                    or isinstance(generation_state, TaskSet)) and (
                        generation_state.is_faulted):
                    generation_state = self.generator.throw(
                        generation_state.exception)
                    continue

                self._reset_timestamp()
                generation_state = self._generate_next(generation_state)

        except StopIteration as sie:
            orchestration_state = OrchestratorState(
                is_done=True,
                output=sie.value,
                actions=self.durable_context.actions,
                custom_status=self.durable_context.custom_status)
        except Exception as e:
            orchestration_state = OrchestratorState(
                is_done=False,
                output=None,  # Should have no output, after generation range
                actions=self.durable_context.actions,
                error=str(e),
                custom_status=self.durable_context.custom_status)

        return orchestration_state.to_json_string()

    def _generate_next(self, partial_result):
        if partial_result is not None:
            gen_result = self.generator.send(partial_result.result)
        else:
            gen_result = self.generator.send(None)
        return gen_result

    def _add_to_actions(self, generation_state):
        if (isinstance(generation_state, Task)
                and hasattr(generation_state, "action")):
            self.durable_context.actions.append([generation_state.action])
        elif (isinstance(generation_state, TaskSet)
              and hasattr(generation_state, "actions")):
            self.durable_context.actions.append(generation_state.actions)

    def _reset_timestamp(self):
        last_timestamp = self.durable_context.decision_started_event.timestamp
        decision_started_events = [e_ for e_ in self.durable_context.histories
                                   if e_.event_type == HistoryEventType.ORCHESTRATOR_STARTED
                                   and e_.timestamp > last_timestamp]
        if len(decision_started_events) == 0:
            self.durable_context.current_utc_datetime = None
        else:
            self.durable_context.decision_started_event = \
                decision_started_events[0]
            self.durable_context.current_utc_datetime = \
                self.durable_context.decision_started_event.timestamp

    @classmethod
    def create(cls, fn: Callable[[DurableOrchestrationContext], Iterator[Any]]) \
            -> Callable[[Any], str]:
        """Create an instance of the orchestration class.

        Parameters
        ----------
        fn: Callable[[DurableOrchestrationContext], Iterator[Any]]
            Generator function that needs orchestration

        Returns
        -------
        Callable[[Any], str]
            Handle function of the newly created orchestration client
        """

        def handle(context: func.OrchestrationContext) -> str:
            context_body = getattr(context, "body", None)
            if context_body is None:
                context_body = context
            return Orchestrator(fn).handle(DurableOrchestrationContext.from_json(context_body))

        return handle



from datetime import datetime
import logging
import os

from urllib.request import urlopen
from PIL import Image
import tensorflow as tf
import numpy as np


scriptpath = os.path.abspath(__file__)
scriptdir = os.path.dirname(scriptpath)
filename = os.path.join(scriptdir, 'model.pb')
labels_filename = os.path.join(scriptdir, 'labels.txt')


output_layer = 'loss:0'
input_node = 'Placeholder:0'

graph_def = tf.GraphDef()
labels = []
network_input_size = 0


def _initialize():
    global labels, network_input_size
    if not labels:
        with tf.io.gfile.GFile(filename, 'rb') as f:
            graph_def.ParseFromString(f.read())
            tf.import_graph_def(graph_def, name='')
        with open(labels_filename, 'rt') as lf:
            labels = [l.strip() for l in lf.readlines()]
        with tf.compat.v1.Session() as sess:
            input_tensor_shape = sess.graph.get_tensor_by_name(
                'Placeholder:0').shape.as_list()
            network_input_size = input_tensor_shape[1]
            logging.info('network_input_size = ' + str(network_input_size))


def _log_msg(msg):
    logging.info("{}: {}".format(datetime.now(), msg))


def _extract_bilinear_pixel(img, x, y, ratio, xOrigin, yOrigin):
    xDelta = (x + 0.5) * ratio - 0.5
    x0 = int(xDelta)
    xDelta -= x0
    x0 += xOrigin
    if x0 < 0:
        x0 = 0
        x1 = 0
        xDelta = 0.0
    elif x0 >= img.shape[1]-1:
        x0 = img.shape[1]-1
        x1 = img.shape[1]-1
        xDelta = 0.0
    else:
        x1 = x0 + 1

    yDelta = (y + 0.5) * ratio - 0.5
    y0 = int(yDelta)
    yDelta -= y0
    y0 += yOrigin
    if y0 < 0:
        y0 = 0
        y1 = 0
        yDelta = 0.0
    elif y0 >= img.shape[0]-1:
        y0 = img.shape[0]-1
        y1 = img.shape[0]-1
        yDelta = 0.0
    else:
        y1 = y0 + 1

    # Get pixels in four corners
    bl = img[y0, x0]
    br = img[y0, x1]
    tl = img[y1, x0]
    tr = img[y1, x1]
    # Calculate interpolation
    b = xDelta * br + (1. - xDelta) * bl
    t = xDelta * tr + (1. - xDelta) * tl
    pixel = yDelta * t + (1. - yDelta) * b
    return pixel.astype(np.uint8)


def _extract_and_resize(img, targetSize):
    determinant = img.shape[1] * targetSize[0] - img.shape[0] * targetSize[1]
    if determinant < 0:
        ratio = float(img.shape[1]) / float(targetSize[1])
        xOrigin = 0
        yOrigin = int(0.5 * (img.shape[0] - ratio * targetSize[0]))
    elif determinant > 0:
        ratio = float(img.shape[0]) / float(targetSize[0])
        xOrigin = int(0.5 * (img.shape[1] - ratio * targetSize[1]))
        yOrigin = 0
    else:
        ratio = float(img.shape[0]) / float(targetSize[0])
        xOrigin = 0
        yOrigin = 0
    resize_image = np.empty(
        (targetSize[0], targetSize[1], img.shape[2]), dtype=np.uint8)
    for y in range(targetSize[0]):
        for x in range(targetSize[1]):
            resize_image[y, x] = _extract_bilinear_pixel(
                img, x, y, ratio, xOrigin, yOrigin)
    return resize_image


def _extract_and_resize_to_256_square(image):
    h, w = image.shape[:2]
    _log_msg("extract_and_resize_to_256_square: " + str(w) + "x" +
             str(h) + " and resize to " + str(256) + "x" + str(256))
    return _extract_and_resize(image, (256, 256))


def _crop_center(img, cropx, cropy):
    h, w = img.shape[:2]
    startx = max(0, w//2-(cropx//2) - 1)
    starty = max(0, h//2-(cropy//2) - 1)
    _log_msg("crop_center: " + str(w) + "x" + str(h) +
             " to " + str(cropx) + "x" + str(cropy))
    return img[starty:starty+cropy, startx:startx+cropx]


def _resize_down_to_1600_max_dim(image):
    w, h = image.size
    if h < 1600 and w < 1600:
        return image

    new_size = (1600 * w // h, 1600) if (h > w) else (1600, 1600 * h // w)
    _log_msg("resize: " + str(w) + "x" + str(h) + " to " +
             str(new_size[0]) + "x" + str(new_size[1]))
    if max(new_size) / max(image.size) >= 0.5:
        method = Image.BILINEAR
    else:
        method = Image.BICUBIC
    return image.resize(new_size, method)


def _convert_to_nparray(image):
    # RGB -> BGR
    _log_msg("Convert to numpy array")
    image = np.array(image)
    return image[:, :, (2, 1, 0)]


def _update_orientation(image):
    exif_orientation_tag = 0x0112
    if hasattr(image, '_getexif'):
        exif = image._getexif()
        if exif != None and exif_orientation_tag in exif:
            orientation = exif.get(exif_orientation_tag, 1)
            _log_msg('Image has EXIF Orientation: ' + str(orientation))
            # orientation is 1 based, shift to zero based and flip/transpose based on 0-based values
            orientation -= 1
            if orientation >= 4:
                image = image.transpose(Image.TRANSPOSE)
            if orientation == 2 or orientation == 3 or orientation == 6 or orientation == 7:
                image = image.transpose(Image.FLIP_TOP_BOTTOM)
            if orientation == 1 or orientation == 2 or orientation == 5 or orientation == 6:
                image = image.transpose(Image.FLIP_LEFT_RIGHT)
    return image


def _predict_image(image):
    if image.mode != "RGB":
        _log_msg("Converting to RGB")
        image.convert("RGB")

    w, h = image.size
    _log_msg("Image size: " + str(w) + "x" + str(h))

    # Update orientation based on EXIF tags
    image = _update_orientation(image)

    # If the image has either w or h greater than 1600 we resize it down respecting
    # aspect ratio such that the largest dimention is 1600
    image = _resize_down_to_1600_max_dim(image)

    # Convert image to numpy array
    image = _convert_to_nparray(image)

    # Crop the center square and resize that square down to 256x256
    resized_image = _extract_and_resize_to_256_square(image)

    # Crop the center for the specified network_input_Size
    cropped_image = _crop_center(
        resized_image, network_input_size, network_input_size)

    tf.compat.v1.reset_default_graph()
    tf.import_graph_def(graph_def, name='')

    with tf.compat.v1.Session() as sess:
        prob_tensor = sess.graph.get_tensor_by_name(output_layer)
        predictions, = sess.run(prob_tensor, {input_node: [cropped_image]})

        result = []
        highest_prediction = None
        for p, label in zip(predictions, labels):
            truncated_probablity = np.float64(round(p, 8))
            if truncated_probablity > 1e-8:
                prediction = {
                    'tagName': label,
                    'probability': truncated_probablity}
                result.append(prediction)
                if not highest_prediction or prediction['probability'] > highest_prediction['probability']:
                    highest_prediction = prediction

        response = {
            'created': datetime.utcnow().isoformat(),
            'predictedTagName': highest_prediction['tagName'],
            'prediction': result
        }

        _log_msg("Results: " + str(response))
        return response


def predict_image_from_url(image_url):
    logging.info("Predicting from url: " + image_url)

    _initialize()

    try:
        with urlopen(image_url) as testImage:
            image = Image.open(testImage)
            return _predict_image(image)
    except Exception as e:
        _log_msg(str(e))
        return None



import logging

from azure.durable_functions import DurableOrchestrationClient
import azure.functions as func


async def main(req: func.HttpRequest, starter: str) -> func.HttpResponse:
    """This function starts up the orchestrator from an HTTP endpoint

    Parameters
    ----------
    req: func.HttpRequest
        An HTTP Request object, it can be used to parse URL
        parameters.

    starter: str
        A JSON-formatted string describing the orchestration context

    Returns
    -------
    func.HttpResponse
        An HTTP response containing useful URLs for monitoring the
        status of newly generated orchestration instance
    """

    logging.debug("Recevied http request call with value {}".format(starter))
    function_name = req.route_params.get('functionName')
    client = DurableOrchestrationClient(starter)

    logging.debug("About to call function {} asyncrounously".format(function_name))
    instance_id = await client.start_new(function_name)

    return client.create_check_status_response(req, instance_id)



import json
from typing import Any

import pytest

from azure.durable_functions.models.OrchestrationRuntimeStatus import OrchestrationRuntimeStatus
from azure.durable_functions.models.DurableOrchestrationClient \
    import DurableOrchestrationClient
from azure.durable_functions.models.DurableOrchestrationStatus import DurableOrchestrationStatus
from tests.conftest import replace_stand_in_bits
from tests.test_utils.constants import RPC_BASE_URL
from unittest.mock import Mock

TEST_INSTANCE_ID = '2e2568e7-a906-43bd-8364-c81733c5891e'
TEST_CREATED_TIME = '2020-01-01T05:00:00Z'
TEST_LAST_UPDATED_TIME = '2020-01-01T05:00:00Z'
MESSAGE_400 = 'instance failed or terminated'
MESSAGE_404 = 'instance not found or pending'
MESSAGE_500 = 'instance failed with unhandled exception'
MESSAGE_501 = "well we didn't expect that"


class MockRequest:
    def __init__(self, expected_url: str, response: [int, any]):
        self._expected_url = expected_url
        self._response = response
        self._get_count = 0

    @property
    def get_count(self):
        return self._get_count

    async def get(self, url: str):
        self._get_count += 1
        assert url == self._expected_url
        return self._response

    async def delete(self, url: str):
        assert url == self._expected_url
        return self._response

    async def post(self, url: str, data: Any = None):
        assert url == self._expected_url
        return self._response


def test_get_start_new_url(binding_string):
    client = DurableOrchestrationClient(binding_string)
    instance_id = "2e2568e7-a906-43bd-8364-c81733c5891e"
    function_name = "my_function"
    start_new_url = client._get_start_new_url(instance_id, function_name)
    expected_url = replace_stand_in_bits(
        f"{RPC_BASE_URL}orchestrators/{function_name}/{instance_id}")
    assert expected_url == start_new_url


def test_get_input_returns_none_when_none_supplied():
    result = DurableOrchestrationClient._get_json_input(None)
    assert result is None


def test_get_input_returns_json_string(binding_string):
    input_ = json.loads(binding_string)
    result = DurableOrchestrationClient._get_json_input(input_)
    input_as_string = json.dumps(input_)
    assert input_as_string == result


def test_get_raise_event_url(binding_string):
    client = DurableOrchestrationClient(binding_string)
    instance_id = "2e2568e7-a906-43bd-8364-c81733c5891e"
    event_name = "test_event_name"
    task_hub_name = "test_task_hub"
    connection_name = "test_connection"
    raise_event_url = client._get_raise_event_url(instance_id, event_name, task_hub_name,
                                                  connection_name)

    expected_url = replace_stand_in_bits(
        f"{RPC_BASE_URL}instances/{instance_id}/raiseEvent/{event_name}"
        f"?taskHub=test_task_hub&connection=test_connection")

    assert expected_url == raise_event_url


def test_create_check_status_response(binding_string):
    client = DurableOrchestrationClient(binding_string)
    instance_id = "2e2568e7-a906-43bd-8364-c81733c5891e"
    request = Mock(url="http://test_azure.net/api/orchestrators/DurableOrchestrationTrigger")
    returned_response = client.create_check_status_response(request, instance_id)

    http_management_payload = {
        "id": instance_id,
        "statusQueryGetUri":
            r"http://test_azure.net/runtime/webhooks/durabletask/instances/"
            r"2e2568e7-a906-43bd-8364-c81733c5891e?taskHub"
            r"=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE",
        "sendEventPostUri":
            r"http://test_azure.net/runtime/webhooks/durabletask/instances/"
            r"2e2568e7-a906-43bd-8364-c81733c5891e/raiseEvent/{"
            r"eventName}?taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE",
        "terminatePostUri":
            r"http://test_azure.net/runtime/webhooks/durabletask/instances/"
            r"2e2568e7-a906-43bd-8364-c81733c5891e/terminate"
            r"?reason={text}&taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE",
        "rewindPostUri":
            r"http://test_azure.net/runtime/webhooks/durabletask/instances/"
            r"2e2568e7-a906-43bd-8364-c81733c5891e/rewind?reason"
            r"={text}&taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE",
        "purgeHistoryDeleteUri":
            r"http://test_azure.net/runtime/webhooks/durabletask/instances/"
            r"2e2568e7-a906-43bd-8364-c81733c5891e"
            r"?taskHub=TASK_HUB_NAME&connection=Storage&code=AUTH_CODE"
    }
    for key, _ in http_management_payload.items():
        http_management_payload[key] = replace_stand_in_bits(http_management_payload[key])
    expected_response = {
        "status_code": 202,
        "body": json.dumps(http_management_payload),
        "headers": {
            "Content-Type": "application/json",
            "Location": http_management_payload["statusQueryGetUri"],
            "Retry-After": "10",
        },
    }

    for k, v in expected_response.get("headers").items():
        assert v == returned_response.headers.get(k)
    assert expected_response.get("status_code") == returned_response.status_code
    assert expected_response.get("body") == returned_response.get_body().decode()


@pytest.mark.asyncio
async def test_get_202_get_status_success(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[202, dict(createdTime=TEST_CREATED_TIME,
                                                   lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                   runtimeStatus="Running")])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status(TEST_INSTANCE_ID)
    assert result is not None
    assert result.runtime_status == "Running"


@pytest.mark.asyncio
async def test_get_200_get_status_success(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, dict(createdTime=TEST_CREATED_TIME,
                                                   lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                   runtimeStatus="Completed")])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status(TEST_INSTANCE_ID)
    assert result is not None
    assert result.runtime_status == "Completed"


@pytest.mark.asyncio
async def test_get_500_get_status_failed(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[500, MESSAGE_500])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status(TEST_INSTANCE_ID)
    assert result is not None
    assert result.message == MESSAGE_500


@pytest.mark.asyncio
async def test_get_400_get_status_failed(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[400, MESSAGE_400])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status(TEST_INSTANCE_ID)
    assert result is not None
    assert result.message == MESSAGE_400


@pytest.mark.asyncio
async def test_get_404_get_status_failed(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[404, MESSAGE_404])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status(TEST_INSTANCE_ID)
    assert result is not None
    assert result.message == MESSAGE_404


@pytest.mark.asyncio
async def test_get_501_get_status_failed(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[501, MESSAGE_501])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    with pytest.raises(Exception):
        await client.get_status(TEST_INSTANCE_ID)


@pytest.mark.asyncio
async def test_get_200_get_status_by_success(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/?runtimeStatus=Running",
                               response=[200, [dict(createdTime=TEST_CREATED_TIME,
                                                    lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                    runtimeStatus="Running"),
                                               dict(createdTime=TEST_CREATED_TIME,
                                                    lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                    runtimeStatus="Running")
                                               ]])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status_by(runtime_status=[OrchestrationRuntimeStatus.Running])
    assert result is not None
    assert len(result) == 2


@pytest.mark.asyncio
async def test_get_500_get_status_by_failed(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/?runtimeStatus=Running",
                               response=[500, MESSAGE_500])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    with pytest.raises(Exception):
        await client.get_status_by(runtime_status=[OrchestrationRuntimeStatus.Running])


@pytest.mark.asyncio
async def test_get_200_get_status_all_success(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/",
                               response=[200, [dict(createdTime=TEST_CREATED_TIME,
                                                    lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                    runtimeStatus="Running"),
                                               dict(createdTime=TEST_CREATED_TIME,
                                                    lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                    runtimeStatus="Running")
                                               ]])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.get_status_all()
    assert result is not None
    assert len(result) == 2


@pytest.mark.asyncio
async def test_get_500_get_status_all_failed(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/",
                               response=[500, MESSAGE_500])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    with pytest.raises(Exception):
        await client.get_status_all()


@pytest.mark.asyncio
async def test_delete_200_purge_instance_history(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, dict(instancesDeleted=1)])
    client = DurableOrchestrationClient(binding_string)
    client._delete_async_request = mock_request.delete

    result = await client.purge_instance_history(TEST_INSTANCE_ID)
    assert result is not None
    assert result.instances_deleted == 1


@pytest.mark.asyncio
async def test_delete_404_purge_instance_history(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[404, MESSAGE_404])
    client = DurableOrchestrationClient(binding_string)
    client._delete_async_request = mock_request.delete

    result = await client.purge_instance_history(TEST_INSTANCE_ID)
    assert result is not None
    assert result.instances_deleted == 0


@pytest.mark.asyncio
async def test_delete_500_purge_instance_history(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[500, MESSAGE_500])
    client = DurableOrchestrationClient(binding_string)
    client._delete_async_request = mock_request.delete

    with pytest.raises(Exception):
        await client.purge_instance_history(TEST_INSTANCE_ID)


@pytest.mark.asyncio
async def test_delete_200_purge_instance_history_by(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/?runtimeStatus=Running",
                               response=[200, dict(instancesDeleted=1)])
    client = DurableOrchestrationClient(binding_string)
    client._delete_async_request = mock_request.delete

    result = await client.purge_instance_history_by(
        runtime_status=[OrchestrationRuntimeStatus.Running])
    assert result is not None
    assert result.instances_deleted == 1


@pytest.mark.asyncio
async def test_delete_404_purge_instance_history_by(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/?runtimeStatus=Running",
                               response=[404, MESSAGE_404])
    client = DurableOrchestrationClient(binding_string)
    client._delete_async_request = mock_request.delete

    result = await client.purge_instance_history_by(
        runtime_status=[OrchestrationRuntimeStatus.Running])
    assert result is not None
    assert result.instances_deleted == 0


@pytest.mark.asyncio
async def test_delete_500_purge_instance_history_by(binding_string):
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/?runtimeStatus=Running",
                               response=[500, MESSAGE_500])
    client = DurableOrchestrationClient(binding_string)
    client._delete_async_request = mock_request.delete

    with pytest.raises(Exception):
        await client.purge_instance_history_by(
            runtime_status=[OrchestrationRuntimeStatus.Running])


@pytest.mark.asyncio
async def test_post_202_terminate(binding_string):
    raw_reason = 'stuff and things'
    reason = 'stuff%20and%20things'
    mock_request = MockRequest(
        expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}/terminate?reason={reason}",
        response=[202, None])
    client = DurableOrchestrationClient(binding_string)
    client._post_async_request = mock_request.post

    result = await client.terminate(TEST_INSTANCE_ID, raw_reason)
    assert result is None


@pytest.mark.asyncio
async def test_post_410_terminate(binding_string):
    raw_reason = 'stuff and things'
    reason = 'stuff%20and%20things'
    mock_request = MockRequest(
        expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}/terminate?reason={reason}",
        response=[410, None])
    client = DurableOrchestrationClient(binding_string)
    client._post_async_request = mock_request.post

    result = await client.terminate(TEST_INSTANCE_ID, raw_reason)
    assert result is None


@pytest.mark.asyncio
async def test_post_404_terminate(binding_string):
    raw_reason = 'stuff and things'
    reason = 'stuff%20and%20things'
    mock_request = MockRequest(
        expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}/terminate?reason={reason}",
        response=[404, MESSAGE_404])
    client = DurableOrchestrationClient(binding_string)
    client._post_async_request = mock_request.post

    with pytest.raises(Exception):
        await client.terminate(TEST_INSTANCE_ID, raw_reason)


@pytest.mark.asyncio
async def test_post_500_terminate(binding_string):
    raw_reason = 'stuff and things'
    reason = 'stuff%20and%20things'
    mock_request = MockRequest(
        expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}/terminate?reason={reason}",
        response=[500, MESSAGE_500])
    client = DurableOrchestrationClient(binding_string)
    client._post_async_request = mock_request.post

    with pytest.raises(Exception):
        await client.terminate(TEST_INSTANCE_ID, raw_reason)


@pytest.mark.asyncio
async def test_wait_or_response_200_completed(binding_string):
    output = 'Some output'
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, dict(createdTime=TEST_CREATED_TIME,
                                                   lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                                                   runtimeStatus="Completed",
                                                   output=output)])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.wait_for_completion_or_create_check_status_response(
        None, TEST_INSTANCE_ID)
    assert result is not None
    assert result.status_code == 200
    assert result.mimetype == 'application/json'
    assert result.get_body().decode() == output


@pytest.mark.asyncio
async def test_wait_or_response_200_canceled(binding_string):
    status = dict(createdTime=TEST_CREATED_TIME,
                  lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                  runtimeStatus="Canceled")
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, status])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.wait_for_completion_or_create_check_status_response(
        None, TEST_INSTANCE_ID)
    assert result is not None
    assert result.status_code == 200
    assert result.mimetype == 'application/json'
    assert json.loads(result.get_body().decode()) == DurableOrchestrationStatus.from_json(
        status).to_json()


@pytest.mark.asyncio
async def test_wait_or_response_200_terminated(binding_string):
    status = dict(createdTime=TEST_CREATED_TIME,
                  lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                  runtimeStatus="Terminated")
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, status])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.wait_for_completion_or_create_check_status_response(
        None, TEST_INSTANCE_ID)
    assert result is not None
    assert result.status_code == 200
    assert result.mimetype == 'application/json'
    assert json.loads(result.get_body().decode()) == DurableOrchestrationStatus.from_json(
        status).to_json()


@pytest.mark.asyncio
async def test_wait_or_response_200_failed(binding_string):
    status = dict(createdTime=TEST_CREATED_TIME,
                  lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                  runtimeStatus="Failed")
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, status])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    result = await client.wait_for_completion_or_create_check_status_response(
        None, TEST_INSTANCE_ID)
    assert result is not None
    assert result.status_code == 500
    assert result.mimetype == 'application/json'
    assert json.loads(result.get_body().decode()) == DurableOrchestrationStatus.from_json(
        status).to_json()


@pytest.mark.asyncio
async def test_wait_or_response_check_status_response(binding_string):
    status = dict(createdTime=TEST_CREATED_TIME,
                  lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                  runtimeStatus="Running")
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, status])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    request = Mock(url="http://test_azure.net/api/orchestrators/DurableOrchestrationTrigger")
    result = await client.wait_for_completion_or_create_check_status_response(
        request, TEST_INSTANCE_ID, timeout_in_milliseconds=2000)
    assert result is not None
    assert mock_request.get_count == 3


@pytest.mark.asyncio
async def test_wait_or_response_check_status_response(binding_string):
    status = dict(createdTime=TEST_CREATED_TIME,
                  lastUpdatedTime=TEST_LAST_UPDATED_TIME,
                  runtimeStatus="Running")
    mock_request = MockRequest(expected_url=f"{RPC_BASE_URL}instances/{TEST_INSTANCE_ID}",
                               response=[200, status])
    client = DurableOrchestrationClient(binding_string)
    client._get_async_request = mock_request.get

    with pytest.raises(Exception):
        await client.wait_for_completion_or_create_check_status_response(
            None, TEST_INSTANCE_ID, timeout_in_milliseconds=500)



import json
import os

import azure.functions as func
import azure.durable_functions as df

def orchestrator_function(context: df.DurableOrchestrationContext):
    """This function provides a sample for activity trigger

    Parameters
    ----------
    context: DurableOrchestrationContext
        This context has the past history and the durable orchestration API

    Returns
    -------
    message
        Returns the result of the activity function return values.

    Yields
    -------
    call_activity: str
        Yields, depending on the `json_rule`, to wait on either all
        tasks to complete, or until one of the tasks completes.
    """

    message = []

    ret_bool = yield context.call_activity("ReturnBool", "1")
    message.append(f"ret_bool: {ret_bool} {type(ret_bool)}")

    # Not supported: return value from activity trigger "bytes" is not json serializable!
    # ret_bytes = yield context.call_activity("ReturnBytes", "1b2b3b")
    # message.append(f"ret_bytes : {ret_bytes} {type(ret_bytes)}")

    ret_dict_of_string = yield context.call_activity("ReturnDictOfString", "kv")
    message.append(f"ret_dict_of_string : {ret_dict_of_string} {type(ret_dict_of_string)}")

    ret_dict_of_string_anno = yield context.call_activity("ReturnDictOfStringWithAnnotation", "kv_anno")
    message.append(f"ret_dict_of_string_anno : {ret_dict_of_string_anno} {type(ret_dict_of_string_anno)}")

    ret_float = yield context.call_activity("ReturnFloat", "123.0")
    message.append(f"ret_float : {ret_float} {type(ret_float)}")

    ret_int = yield context.call_activity("ReturnInt", "123")
    message.append(f"ret_int : {ret_int} {type(ret_int)}")

    ret_int_from_float = yield context.call_activity("ReturnIntFromFloat", 3.14)
    message.append(f"ret_int_from_float : {ret_int_from_float} {type(ret_int_from_float)}")

    ret_list_of_float = yield context.call_activity("ReturnListOfFloat", "4.5")
    message.append(f"ret_list_of_float : {ret_list_of_float} {type(ret_list_of_float)}")

    ret_list_of_float_anno = yield context.call_activity("ReturnListOfFloatWithAnnotation", "5.6")
    message.append(f"ret_list_of_float_anno : {ret_list_of_float_anno} {type(ret_list_of_float_anno)}")

    # Not supported: return value from activity trigger "set" is not json serializable!
    # ret_set_of_int = yield context.call_activity("ReturnSetOfInt", 5)
    # message.append(f"ret_set_of_int : {ret_set_of_int} {type(ret_set_of_int)}")

    ret_string = yield context.call_activity("ReturnString", "simple_string")
    message.append(f"ret_string : {ret_string} {type(ret_string)}")

    return message


main = df.Orchestrator.create(orchestrator_function)



from uuid import uuid1
import datetime
from typing import List, Any, Dict
from datetime import datetime

from azure.durable_functions.tasks.new_uuid import URL_NAMESPACE, \
    _create_deterministic_uuid
from azure.durable_functions.models.DurableOrchestrationContext import DurableOrchestrationContext
from azure.durable_functions.constants import DATETIME_STRING_FORMAT


def test_create_deterministic_uuid():
    namespace = URL_NAMESPACE
    instance_id = uuid1()
    current_utc_datetime = datetime.now().strftime(DATETIME_STRING_FORMAT);

    name1 = f"{instance_id}_{current_utc_datetime}_0"
    name2 = f"{instance_id}_{current_utc_datetime}_12"

    result1a = _create_deterministic_uuid(namespace, name1)
    result1b = _create_deterministic_uuid(namespace, name1)

    result2a = _create_deterministic_uuid(namespace, name2)
    result2b = _create_deterministic_uuid(namespace, name2)

    assert result1a == result1b
    assert result2a == result2b

    assert result1a != result2a
    assert result1b != result2b


def history_list() -> List[Dict[Any, Any]]:
    history = [{'EventType': 12, 'EventId': -1, 'IsPlayed': False,
                'Timestamp': '2019-12-08T23:18:41.3240927Z'}, {
                   'OrchestrationInstance': {'InstanceId': '48d0f95957504c2fa579e810a390b938',
                                             'ExecutionId': 'fd183ee02e4b4fd18c95b773cfb5452b'},
                   'EventType': 0, 'ParentInstance': None, 'Name': 'DurableOrchestratorTrigger',
                   'Version': '', 'Input': 'null', 'Tags': None, 'EventId': -1, 'IsPlayed': False,
                   'Timestamp': '2019-12-08T23:18:39.756132Z'}]
    return history


def test_new_uuid():
    instance_id = str(uuid1())
    history = history_list()
    context1 = DurableOrchestrationContext(history, instance_id, False, None)

    result1a = context1.new_uuid()
    result1b = context1.new_uuid()

    context2 = DurableOrchestrationContext(history, instance_id, False, None)

    result2a = context2.new_uuid()
    result2b = context2.new_uuid()

    assert result1a == result2a
    assert result1b == result2b

    assert result1a != result1b
    assert result2a != result2b



from ..models.TaskSet import TaskSet


def task_any(tasks):
    """Determine whether any of the given tasks is completed.

    Parameters
    ----------
    tasks : Task
        The tasks to evaluate their current state.

    Returns
    -------
    TaskSet
        Returns a completed Durable Task Set if any of the tasks is completed.
        Returns a not completed Durable Task Set if none of the tasks are completed.
        Returns a faulted Taskset if all tasks are faulted
    """
    all_actions = []
    completed_tasks = []
    faulted_tasks = []
    error_message = []
    for task in tasks:
        if isinstance(task, TaskSet):
            for action in task.actions:
                all_actions.append(action)
        else:
            all_actions.append(task.action)

        if task.is_faulted:
            faulted_tasks.append(task)
            error_message.append(task.exception)
        elif task.is_completed:
            completed_tasks.append(task)

    completed_tasks.sort(key=lambda t: t.timestamp)

    if len(faulted_tasks) == len(tasks):
        return TaskSet(True, all_actions, None, is_faulted=True, exception=Exception(
            f"All tasks have failed, errors messages in all tasks:{error_message}"))
    elif len(completed_tasks) != 0:
        return TaskSet(True, all_actions, completed_tasks[0], False, completed_tasks[0].timestamp)
    else:
        return TaskSet(False, all_actions, None)



from typing import List

from ..models.Task import (
    Task)
from ..models.actions.WaitForExternalEventAction import WaitForExternalEventAction
from ..models.history import HistoryEvent
from .task_utilities import set_processed, parse_history_event, find_event_raised


def wait_for_external_event_task(
        state: List[HistoryEvent],
        name: str) -> Task:
    """Determine the state of a task that is waiting for an event to occur.

    Parameters
    ----------
    state : List[HistoryEvent]
        The list of history events to search to determine the current
    state of the task.
    name : str
        The event name of the event that the task is waiting for.

    Returns
    -------
    Task
        Returns a completed task if the expected event was raised.
        Returns a not completed task if the expected event has not occurred yet.
    """
    new_action = WaitForExternalEventAction(name)
    event_raised = find_event_raised(state, name)
    set_processed([event_raised])
    if event_raised:
        return Task(
            is_completed=True,
            is_faulted=False,
            action=new_action,
            result=parse_history_event(event_raised),
            timestamp=event_raised.timestamp,
            id_=event_raised.event_id)

    else:
        return Task(is_completed=False, is_faulted=False, action=new_action)



from .orchestrator_test_utils \
    import get_orchestration_state_result, assert_orchestration_state_equals, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.RetryOptions import RetryOptions
from azure.durable_functions.models.actions.CallActivityWithRetryAction \
    import CallActivityWithRetryAction


RETRY_OPTIONS = RetryOptions(5000, 3)


def generator_function(context):
    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = yield context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    task2 = yield context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")
    task3 = yield context.call_activity_with_retry(
        "Hello",  retry_options, "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs


def base_expected_state(output=None) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output)


def add_hello_action(state: OrchestratorState, input_: str):
    retry_options = RETRY_OPTIONS
    action = CallActivityWithRetryAction(
        function_name='Hello', retry_options=retry_options, input_=input_)
    state._actions.append([action])


def add_hello_failed_events(
        context_builder: ContextBuilder, id_: int, reason: str, details: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_failed_event(
        id_=id_, reason=reason, details=details)


def add_hello_completed_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=id_, result=result)


def add_retry_timer_events(context_builder: ContextBuilder, id_: int):
    fire_at = context_builder.add_timer_created_event(id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_timer_fired_event(id_=id_, fire_at=fire_at)


def test_initial_orchestration_state():
    context_builder = ContextBuilder('test_simple_function')

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_retry():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_timer_entry():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_failed_retry():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_failed_retry_timer_added():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_successful_tokyo_with_failed_retry_timer_added():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_completed_events(context_builder, 2, "\"Hello Tokyo!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_hit_max_attempts():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_hello_failed_events(context_builder, 4, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 5)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected_state._error = f'{failed_reason} \n {failed_details}'
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)



import nox

@nox.session(python="3.7")
def tests(session):
    # same as pip install -r -requirements.txt
    session.install("-r", "requirements.txt")
    session.install("pytest")
    session.run("pytest", "-v", "tests")


@nox.session(python="3.7")
def lint(session):
    session.install("flake8")
    session.install("flake8-docstrings")
    session.run("flake8", "./azure/")



import json
import os
from azure.cognitiveservices.search.imagesearch import ImageSearchClient
from msrest.authentication import CognitiveServicesCredentials


def _get_cognitive_services_client() -> ImageSearchClient:
    """Get the cognitive service client to run the searches against.

    Ensure there is a COGNITIVE_KEY and COGNITIVE_ENDPOINT configured in your
    app setting for the function, or your local.settings.json file when running
    locally.

    Returns
    -------
    client: ImageSearchClient
        Cognitive service client
    """
    subscription_key = os.environ.get('COGNITIVE_KEY')
    subscription_endpoint = os.environ.get('COGNITIVE_ENDPOINT')
    client = ImageSearchClient(endpoint=subscription_endpoint,
                               credentials=CognitiveServicesCredentials(subscription_key))
    return client


def main(value: str) -> str:
    """Get a list of image URLs from Bing Search to run predictions against.

    Parameters
    ----------
    value: str
        The number of images to get

    Returns
    -------
    str
        List of image URLs to run the prediction against
    """
    client = _get_cognitive_services_client()

    volume_of_images = int(value)
    increment = volume_of_images if volume_of_images < 100 else 100
    image_urls = []
    offset = 0
    search_term = "dog OR cat"

    # search cognitive services until we have the volume of image URLs requested
    while len(image_urls) < volume_of_images:
        search_results = client.images.search(
            query=search_term, count=increment, offset=offset)
        image_urls.extend(
            [image.content_url for image in search_results.value])
        offset += increment
        increment = increment if offset + \
            increment < volume_of_images else volume_of_images - offset

    return json.dumps(image_urls)



def main(value: str) -> str:
    """Activity function to validate that a number is within range

    Parameters
    ----------
    value: str
        A number value, expected to be lesser than 6

    Returns
    -------
    value: str
        The input value, assuming it was lesser than 6
    """
    int_value = int(value)
    if int_value >= 6:
      raise Exception('Bad Request')

    return value


from typing import List, Any

from .task_utilities import find_task_scheduled, \
    find_task_retry_timer_created, set_processed, parse_history_event, \
    find_task_completed, find_task_failed, find_task_retry_timer_fired
from ..models.RetryOptions import RetryOptions
from ..models.Task import (
    Task)
from ..models.actions.CallActivityWithRetryAction import \
    CallActivityWithRetryAction
from ..models.history import HistoryEvent


def call_activity_with_retry_task(
        state: List[HistoryEvent],
        retry_options: RetryOptions,
        name: str,
        input_: Any = None) -> Task:
    """Determine the state of scheduling an activity for execution with retry options.

    Parameters
    ----------
    state: List[HistoryEvent]
        The list of history events to search to determine the current state of the activity.
    retry_options: RetryOptions
        The retry options for the activity function.
    name: str
        The name of the activity function to call.
    input_: Any
        The JSON-serializable input to pass to the activity function.

    Returns
    -------
    Task
        A Durable Task that completes when the called activity function completes or fails
        completely.
    """
    new_action = CallActivityWithRetryAction(
        function_name=name, retry_options=retry_options, input_=input_)
    for attempt in range(retry_options.max_number_of_attempts):
        task_scheduled = find_task_scheduled(state, name)
        task_completed = find_task_completed(state, task_scheduled)
        task_failed = find_task_failed(state, task_scheduled)
        task_retry_timer = find_task_retry_timer_created(state, task_failed)
        task_retry_timer_fired = find_task_retry_timer_fired(
            state, task_retry_timer)
        set_processed([task_scheduled, task_completed,
                       task_failed, task_retry_timer, task_retry_timer_fired])

        if not task_scheduled:
            break

        if task_completed:
            return Task(
                is_completed=True,
                is_faulted=False,
                action=new_action,
                result=parse_history_event(task_completed),
                timestamp=task_completed.timestamp,
                id_=task_completed.TaskScheduledId)

        if task_failed and task_retry_timer and attempt + 1 >= \
                retry_options.max_number_of_attempts:
            return Task(
                is_completed=True,
                is_faulted=True,
                action=new_action,
                timestamp=task_failed.timestamp,
                id_=task_failed.TaskScheduledId,
                exc=Exception(
                    f"{task_failed.Reason} \n {task_failed.Details}")
            )

    return Task(is_completed=False, is_faulted=False, action=new_action)
